"""
Showcases how a HDC-host can communicate with a HDC-device which is running the Demo_Minimal firmware example.
"""
import logging
import time

from minimal_proxy import MinimalDevice, MinimalCore


def showcase_minimal():
    #########################################################
    # This example uses python logging to explain the
    # demonstration, but also show internals of the HDC-host
    demo_logger = logging.getLogger('showcase_minimal')
    demo_logger.setLevel(logging.INFO)
    #
    hdc_root_logger = logging.getLogger()
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter('%(asctime)s.%(msecs)03d - %(levelname)7s - %(name)s - %(message)s',
                                               datefmt='%M:%S'))
    hdc_root_logger.addHandler(log_handler)

    # You can tweak the following log-levels to tune verbosity of HDC internals:
    logging.getLogger("hdcproto.transport.packetizer").setLevel(logging.WARNING)
    logging.getLogger("hdcproto.transport.serialport").setLevel(logging.WARNING)
    logging.getLogger("hdcproto.host.router").setLevel(logging.WARNING)
    logging.getLogger("hdcproto.host.proxy").setLevel(logging.NOTSET)

    #################################################
    # Connect to HDC-device at a specific serial port
    dev = MinimalDevice(connection_url="COM10")  # Note how this implements all HDC specifics of a given device type
    dev.router.connect()  # Will fail if your device is connected at a different port.

    ######################################################################################
    # Example of how "inheritance" mixes-in the stuff defined in DeviceProxyBase into self
    demo_logger.info("--------------------------------------------------------------")
    demo_logger.info("Asking device about the HDC-spec version it is compliant with:")
    demo_logger.info(f"Raw version string: '{dev.get_hdc_version_string()}'")
    demo_logger.info(f"    Parsed version: '{repr(dev.get_hdc_version())}'")

    ########################################################################
    # Example of how to react to an event on the very moment it is received.
    # WARNING: This handler will be executed in the "receiver thread"
    #          and thus must be fast and thread-safe!
    #          Note how the HDC-host driver is neither, thus refrain from
    #          using the HDC-host driver from within this kind of handler!
    #
    #          See example further below on how to react to events in a delayed, but much safer manner.
    def button_event_handler(event_payload: MinimalCore.ButtonEventPayload):
        demo_logger.info(f"ButtonID:0x{event_payload.button_id:02X} ButtonState:{event_payload.button_state}")

    dev.core.evt_button.register_event_payload_handler(button_event_handler)

    ##################################################################
    # Example of how the host requests the device to execute a command
    # This is essentially a "remote procedure call".
    # Resetting the device will raise some StateTransition-events.
    # ToDo: A more instructive example with arguments and return value.
    demo_logger.info("_____________________________")
    demo_logger.info("Resetting the Core-feature...")
    dev.core.cmd_reset()  # Blocks until it receives reply from HDC-device or the default timeout elapses.
    time.sleep(0.5)  # Allow for some time for the actual firmware reset to happen.

    ################################################################################################
    # Example of how "composition" keeps stuff defined in CoreFeatureProxyBase separate in self.hdc
    demo_logger.info("_____________________________________")
    demo_logger.info("Obtain some mandatory property values...")
    demo_logger.info(f"       LogEventThreshold: {dev.core.hdc.prop_log_event_threshold.get_value_name()}")

    ##################################################################
    # Example of how the host gets property values
    demo_logger.info("_____________________________________")
    demo_logger.info("Obtain some custom property values...")
    demo_logger.info(f"   Microcontroller REVID: 0x{dev.core.prop_microcontroller_revid.get():08x}")
    demo_logger.info(f"   Microcontroller DEVID: 0x{dev.core.prop_microcontroller_devid.get():08x}")
    demo_logger.info(f"   Microcontroller   UID: 0x{dev.core.prop_microcontroller_uid.get().hex()}")

    demo_logger.info("_____________________________________________________________________________")
    demo_logger.info("Change LED blinking rate, depending on most recently received button event...")
    try:
        while True:

            time.sleep(2)

            #######################################################################################################
            # Example of how a host can process events at a more suitable point in time and a more suitable thread.
            event_payload_deque = dev.core.evt_button.most_recently_received_event_payloads  # Just for readability
            if not event_payload_deque:
                continue  # No button-events have been received, thus we can skip the remainder.

            most_recent_button_event: MinimalCore.ButtonEventPayload = event_payload_deque.pop()
            event_payload_deque.clear()  # Get rid of any other, prior events.

            ##############################################
            # Example of how the host sets property values
            new_led_blinking_rate = 5 if most_recent_button_event.button_state else 20
            dev.core.prop_led_blinking_rate.set(new_led_blinking_rate)
    finally:
        dev.router.close()


if __name__ == '__main__':
    showcase_minimal()



import typing
import unittest

from hdcproto.common import HdcDataType


class TestPayloadParser(unittest.TestCase):

    def test_expecting_void_and_getting_empty_payload(self):
        empty_payload = bytes()

        self.assertIsNone(
            HdcDataType.parse_payload(
                raw_payload=empty_payload,
                expected_data_types=None)
        )

        self.assertIsNone(
            HdcDataType.parse_payload(
                raw_payload=empty_payload,
                expected_data_types=[])
        )

        # noinspection PyTypeChecker
        self.assertIsNone(
            HdcDataType.parse_payload(
                raw_payload=empty_payload,
                expected_data_types=[None])  # Very weird, I know. But why not?
        )

    def test_expecting_void_but_getting_non_empty_payload(self):
        non_empty_payload = bytes(range(1))

        with self.assertRaises(ValueError):
            HdcDataType.parse_payload(
                raw_payload=non_empty_payload,
                expected_data_types=None
            )

        with self.assertRaises(ValueError):
            HdcDataType.parse_payload(
                raw_payload=non_empty_payload,
                expected_data_types=[]
            )

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            HdcDataType.parse_payload(
                raw_payload=non_empty_payload,
                expected_data_types=[None]  # Very weird, I know. But why not?
            )

    def test_validation_of_expected_type_argument(self):
        uint8_payload = bytes(range(1))

        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            HdcDataType.parse_payload(
                raw_payload=uint8_payload,
                expected_data_types=int
            )

        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            HdcDataType.parse_payload(
                raw_payload=uint8_payload,
                expected_data_types=[int]
            )

        with self.assertRaises(TypeError):
            HdcDataType.parse_payload(
                raw_payload=uint8_payload,
                expected_data_types=[HdcDataType.UINT8, int]
            )

    def test_result_as_list_or_scalar(self):
        uint8_value = 0x42
        uint8_payload = HdcDataType.UINT8.value_to_bytes(uint8_value)

        self.assertIsInstance(
            HdcDataType.parse_payload(
                raw_payload=uint8_payload,
                expected_data_types=HdcDataType.UINT8  # Not a list...
            ),
            int  # ... thus expecting a scalar as result!
        )

        self.assertIsInstance(
            HdcDataType.parse_payload(
                raw_payload=uint8_payload,
                expected_data_types=[HdcDataType.UINT8]  # A list...
            ),
            list  # ... thus expecting a list as result!
        )

    def test_parsing_of_uint8(self):
        uint8_value = 0x42
        uint8_payload = HdcDataType.UINT8.value_to_bytes(uint8_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=uint8_payload, expected_data_types=HdcDataType.UINT8),
            uint8_value
        )

    def test_parsing_of_uint16(self):
        uint16_value = 0x4242
        uint16_payload = HdcDataType.UINT16.value_to_bytes(uint16_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=uint16_payload, expected_data_types=HdcDataType.UINT16),
            uint16_value
        )

    def test_parsing_of_uint32(self):
        uint32_value = 0x42424242
        uint32_payload = HdcDataType.UINT32.value_to_bytes(uint32_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=uint32_payload, expected_data_types=HdcDataType.UINT32),
            uint32_value
        )

    def test_parsing_of_int8(self):
        int8_value = -0x42
        int8_payload = HdcDataType.INT8.value_to_bytes(int8_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=int8_payload, expected_data_types=HdcDataType.INT8),
            int8_value
        )

    def test_parsing_of_int16(self):
        int16_value = -0x4242
        int16_payload = HdcDataType.INT16.value_to_bytes(int16_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=int16_payload, expected_data_types=HdcDataType.INT16),
            int16_value
        )

    def test_parsing_of_int32(self):
        int32_value = -0x42424242
        int32_payload = HdcDataType.INT32.value_to_bytes(int32_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=int32_payload, expected_data_types=HdcDataType.INT32),
            int32_value
        )

    def test_parsing_of_float(self):
        float_value = -42.424242
        float_payload = HdcDataType.FLOAT.value_to_bytes(float_value)

        self.assertAlmostEqual(
            HdcDataType.parse_payload(raw_payload=float_payload, expected_data_types=HdcDataType.FLOAT),
            float_value,
            places=5  # Note the limited numeric precision of FLOAT !!!
        )

    def test_parsing_of_double(self):
        double_value = -42.424242
        double_payload = HdcDataType.DOUBLE.value_to_bytes(double_value)

        self.assertAlmostEqual(
            HdcDataType.parse_payload(raw_payload=double_payload, expected_data_types=HdcDataType.DOUBLE),
            double_value,
            places=6  # Note the better numeric precision of DOUBLE
        )

    def test_parsing_of_bool_true(self):
        true_value = True
        true_payload = HdcDataType.BOOL.value_to_bytes(true_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=true_payload, expected_data_types=HdcDataType.BOOL),
            true_value
        )

    def test_parsing_of_bool_false(self):
        false_value = False
        false_payload = HdcDataType.BOOL.value_to_bytes(false_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=false_payload, expected_data_types=HdcDataType.BOOL),
            false_value
        )

    def test_parsing_of_nonempty_blob(self):
        blob_value = bytes(range(42))
        blob_payload = HdcDataType.BLOB.value_to_bytes(blob_value)

        self.assertEqual(blob_value, blob_payload)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=blob_payload, expected_data_types=HdcDataType.BLOB),
            blob_value
        )

    def test_parsing_of_empty_blob(self):
        blob_value = bytes(range(0))
        blob_payload = HdcDataType.BLOB.value_to_bytes(blob_value)

        self.assertEqual(blob_value, blob_payload)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=blob_payload, expected_data_types=HdcDataType.BLOB),
            blob_value
        )

    def test_parsing_of_nonempty_utf8(self):
        # noinspection SpellCheckingInspection
        utf8_value = "Lorem ipsum ùոïċọɗẹ"
        utf8_payload = HdcDataType.UTF8.value_to_bytes(utf8_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=utf8_payload, expected_data_types=HdcDataType.UTF8),
            utf8_value
        )

    def test_parsing_of_empty_utf8(self):
        utf8_value = ""
        utf8_payload = HdcDataType.UTF8.value_to_bytes(utf8_value)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=utf8_payload, expected_data_types=HdcDataType.UTF8),
            utf8_value
        )

    def test_parsing_multiple_values(self):  # Without any variable size types!
        types_and_values: list[tuple[HdcDataType, typing.Any]] = [
            (HdcDataType.UINT8, 0x42),
            (HdcDataType.INT32, -0x42424242),
            (HdcDataType.DOUBLE, -42.424242),
            (HdcDataType.BOOL, True)
        ]

        payload = bytearray()
        multiple_values = list()
        expected_data_types = list()
        for datatype, value in types_and_values:
            payload.extend(datatype.value_to_bytes(value))
            multiple_values.append(value)
            expected_data_types.append(datatype)
        payload = bytes(payload)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=payload, expected_data_types=expected_data_types),
            multiple_values
        )

    def test_parsing_multiple_values_with_longer_payload_than_expected(self):
        types_and_values: list[tuple[HdcDataType, typing.Any]] = [
            (HdcDataType.UINT8, 0x42),
            (HdcDataType.INT32, -0x42424242),
            (HdcDataType.DOUBLE, -42.424242),
            (HdcDataType.BOOL, True)
        ]

        payload = bytearray()
        multiple_values = list()
        expected_data_types = list()
        for datatype, value in types_and_values:
            payload.extend(datatype.value_to_bytes(value))
            multiple_values.append(value)
            expected_data_types.append(datatype)

        payload.append(0x42)  # This is the one bogus byte that the parser should be complaining about!
        payload = bytes(payload)

        with self.assertRaises(ValueError):
            HdcDataType.parse_payload(raw_payload=payload, expected_data_types=expected_data_types)

    def test_parsing_multiple_values_with_shorter_payload_than_expected(self):
        types_and_values: list[tuple[HdcDataType, typing.Any]] = [
            (HdcDataType.UINT8, 0x42),
            (HdcDataType.INT32, -0x42424242),
            (HdcDataType.DOUBLE, -42.424242),
            (HdcDataType.BOOL, True)
        ]

        payload = bytearray()
        multiple_values = list()
        expected_data_types = list()
        for datatype, value in types_and_values:
            payload.extend(datatype.value_to_bytes(value))
            multiple_values.append(value)
            expected_data_types.append(datatype)

        payload = payload[:-1]  # This is the one missing byte that the parser should be complaining about!
        payload = bytes(payload)

        with self.assertRaises(ValueError):
            HdcDataType.parse_payload(raw_payload=payload, expected_data_types=expected_data_types)

    def test_parsing_multiple_values_and_last_is_of_variable_size(self):
        # noinspection SpellCheckingInspection
        types_and_values: list[tuple[HdcDataType, typing.Any]] = [
            (HdcDataType.UINT8, 0x42),
            (HdcDataType.INT32, -0x42424242),
            (HdcDataType.DOUBLE, -42.424242),
            (HdcDataType.BOOL, True),
            (HdcDataType.UTF8, "Lorem ipsum ùոïċọɗẹ")
        ]

        payload = bytearray()
        multiple_values = list()
        expected_data_types = list()
        for datatype, value in types_and_values:
            payload.extend(datatype.value_to_bytes(value))
            multiple_values.append(value)
            expected_data_types.append(datatype)
        payload = bytes(payload)

        self.assertEqual(
            HdcDataType.parse_payload(raw_payload=payload, expected_data_types=expected_data_types),
            multiple_values
        )

    def test_parsing_multiple_values_and_first_is_of_variable_size(self):
        # noinspection SpellCheckingInspection
        types_and_values: list[tuple[HdcDataType, typing.Any]] = [
            (HdcDataType.UTF8, "Lorem ipsum ùոïċọɗẹ"),
            (HdcDataType.UINT8, 0x42),
        ]

        payload = bytearray()
        multiple_values = list()
        expected_data_types = list()
        for datatype, value in types_and_values:
            payload.extend(datatype.value_to_bytes(value))
            multiple_values.append(value)
            expected_data_types.append(datatype)
        payload = bytes(payload)

        with self.assertRaises(ValueError):
            # Should fail, because variable size can only be inferred for the last argument!
            HdcDataType.parse_payload(raw_payload=payload, expected_data_types=expected_data_types)


if __name__ == '__main__':
    unittest.main()



"""
Proxy classes for the device and its features and properties
"""
from __future__ import annotations

import collections
import enum
import logging
import semver
import time
import typing
from datetime import datetime

import hdcproto.host.router
import hdcproto.transport.serialport
from hdcproto.common import (HdcError, MessageTypeID, FeatureID, CmdID, ReplyErrorCode, EvtID, PropID, HdcDataType,
                             is_valid_uint8)

logger = logging.getLogger(__name__)  # Logger-name: "hdcproto.host.proxy"

DEFAULT_REPLY_TIMEOUT = 0.2


class CommandProxyBase:
    feature_proxy: FeatureProxyBase
    command_id: int
    default_timeout: float
    known_errors: dict[int, str]
    msg_prefix: bytes

    def __init__(self,
                 feature_proxy: FeatureProxyBase,
                 command_id: int,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):

        # Looks like an instance-attribute, but it's more of a class-attribute, actually. ;-)
        # Logger-name like: "hdcproto.host.proxy.MyDeviceProxy.MyFeatureProxy.MyCommandProxy"
        self.logger = feature_proxy.logger.getChild(self.__class__.__name__)

        self.logger.debug(f"Initializing instance of {self.__class__.__name__} "
                          f"as proxy for CommandID=0x{command_id:02X} on FeatureID=0x{feature_proxy.feature_id:02X}")

        if not is_valid_uint8(command_id):
            raise ValueError(f"command_id value of {command_id} is beyond valid range from 0x00 to 0xFF")

        self.feature_proxy = feature_proxy
        self.command_id = command_id
        self.default_timeout = default_timeout
        self.known_errors = dict()

        self.msg_prefix = bytes([int(MessageTypeID.COMMAND),
                                 self.feature_proxy.feature_id,
                                 self.command_id])

        # Register mandatory error codes
        self.register_error(ReplyErrorCode.NO_ERROR)
        self.register_error(ReplyErrorCode.UNKNOWN_FEATURE)
        self.register_error(ReplyErrorCode.UNKNOWN_COMMAND)
        self.register_error(ReplyErrorCode.INCORRECT_COMMAND_ARGUMENTS)
        self.register_error(ReplyErrorCode.COMMAND_NOT_ALLOWED_NOW)
        self.register_error(ReplyErrorCode.COMMAND_FAILED)

    def register_error(self, code: int | ReplyErrorCode, error_name: str | None = None) -> None:
        if isinstance(code, ReplyErrorCode):
            code = int(code)
            error_name = str(code)
        else:
            # Disallow overriding ReplyErrorCodes whose meaning is already predefined and thus reserved by HDC-spec
            try:
                code_as_defined_by_hdc_spec = ReplyErrorCode(code)
            except ValueError:
                pass  # Meaning this code is *not* defined by HDC-spec, thus it's available for custom use.
            else:
                raise ValueError(f"Failed to register ReplyErrorCode 0x{code:02X}:'{error_name}', because that it's "
                                 f"already defined by HDC-spec to mean '{code_as_defined_by_hdc_spec}'")

        code = int(code)

        if not is_valid_uint8(code):
            raise ValueError(f"Reply error code of {code} is beyond valid range from 0x00 to 0xFF")

        if error_name is None:
            error_name = f"Error 0x{code:02X}"  # Fallback for lazy callers

        if code in self.known_errors:
            raise ValueError(f'Already registered ErrorCode {code} as "{self.known_errors[code]}"')
        self.known_errors[code] = error_name

    def _send_request_and_get_reply(self, request_message: bytes, timeout: float) -> bytes:

        if not request_message.startswith(self.msg_prefix):
            raise ValueError("Request does not match the expected prefix")

        p = self.feature_proxy.router_feature.router  # Just an alias for readability

        reply_message = p.send_request_and_get_reply(request_message, timeout)

        if not reply_message.startswith(self.msg_prefix):
            # ToDo: Might be a delayed reply to a previous request that timed-out. Maybe we should just ignore it?
            raise HdcError("Reply does not match the expected prefix")

        if len(reply_message) < 4:
            raise HdcError(f"Expected a reply size of 4 or more bytes, "
                           f"but received {len(reply_message)}")

        error_code = reply_message[3]
        if error_code != ReplyErrorCode.NO_ERROR:
            error_name = self.known_errors.get(error_code, f"Unknown error code 0x{error_code:02X}")
            raise HdcReplyError(error_name, reply_message)

        return reply_message

    def _call_cmd(self,
                  cmd_args: list[tuple[HdcDataType, bool | int | float | str | bytes]] | None,
                  return_types: HdcDataType | list[HdcDataType] | None,
                  timeout: float | None = None) -> typing.Any:
        request_message = bytearray(self.msg_prefix)

        if cmd_args is None:
            cmd_args = list()

        for arg_data_type, arg_value in cmd_args:
            arg_as_raw_bytes = arg_data_type.value_to_bytes(arg_value)
            request_message.extend(arg_as_raw_bytes)

        if timeout is None:
            timeout = self.default_timeout

        self.logger.info(f"Executing CommandID=0x{self.command_id:02X}")
        reply_message = self._send_request_and_get_reply(request_message=request_message,
                                                         timeout=timeout)
        self.logger.debug(f"Finished executing CommandID=0x{self.command_id:02X}")

        try:
            return_values = HdcDataType.parse_reply_msg(reply_message=reply_message,
                                                        expected_data_types=return_types)
        except ValueError as e:
            raise HdcError(f"Failed to parse reply to CommandID={self.command_id:02X}, because: {e}")

        if isinstance(return_values, list):
            # Be more pythonic by returning a tuple, instead of a list
            return_values = tuple(return_values)

        return return_values


class VoidWithoutArgsCommandProxy(CommandProxyBase):
    """Convenience proxy-class for FeatureCommands that neither have arguments nor return values."""

    def __init__(self, feature_proxy: FeatureProxyBase, command_id: int, default_timeout: float | None = None):
        super().__init__(feature_proxy, command_id=command_id, default_timeout=default_timeout)

    def __call__(self, timeout: float | None = None):
        return super()._call_cmd(cmd_args=None, return_types=None, timeout=timeout)


class GetPropertyNameCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_PROP_NAME)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)

    def __call__(self, property_id: int, timeout: float | None = None) -> str:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, property_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class GetPropertyTypeCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_PROP_TYPE)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)

    def __call__(self, property_id: int, timeout: float | None = None) -> HdcDataType:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        property_type_id = super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, property_id), ],
            return_types=HdcDataType.UINT8,
            timeout=timeout
        )
        property_type = HdcDataType(property_type_id)
        return property_type


class GetPropertyReadonlyCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_PROP_RO)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)

    def __call__(self, property_id: int, timeout: float | None = None) -> bool:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, property_id), ],
            return_types=HdcDataType.BOOL,
            timeout=timeout
        )


class GetPropertyValueCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_PROP_VALUE)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)

    def __call__(self,
                 property_id: int,
                 property_data_type: HdcDataType,
                 timeout: float | None = None) -> typing.Any:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, property_id), ],
            return_types=property_data_type,
            timeout=timeout
        )


class SetPropertyValueCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.SET_PROP_VALUE)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)
        self.register_error(ReplyErrorCode.INVALID_PROPERTY_VALUE)
        self.register_error(ReplyErrorCode.PROPERTY_IS_READ_ONLY)

    def __call__(self,
                 property_id: int,
                 property_data_type: HdcDataType,
                 new_value: bool | int | float | str | bytes,
                 timeout: float | None = None) -> bool | int | float | str | bytes:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        return super()._call_cmd(
            cmd_args=[
                (HdcDataType.UINT8, property_id),
                (property_data_type, new_value),
            ],
            return_types=property_data_type,
            timeout=timeout
        )


class GetPropertyDescriptionCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_PROP_DESCR)
        self.register_error(ReplyErrorCode.UNKNOWN_PROPERTY)

    def __call__(self, property_id: int, timeout: float | None = None) -> str:
        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, property_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class GetCommandNameCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_CMD_NAME)
        # Reuses ReplyErrorCode.UNKNOWN_COMMAND to also mean we requested the name for an unknown CommandID

    def __call__(self, command_id: int, timeout: float | None = None) -> str:
        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, command_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class GetCommandDescriptionCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_CMD_DESCR)
        # Reuses ReplyErrorCode.UNKNOWN_COMMAND to also mean we requested the description for an unknown CommandID

    def __call__(self, command_id: int, timeout: float | None = None) -> str:
        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, command_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class GetEventNameCommandProxy(CommandProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_EVT_NAME)
        self.register_error(ReplyErrorCode.UNKNOWN_EVENT)

    def __call__(self, event_id: int, timeout: float | None = None) -> str:
        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, event_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class GetEventDescriptionCommandProxy(CommandProxyBase):
    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, command_id=CmdID.GET_EVT_DESCR)
        self.register_error(ReplyErrorCode.UNKNOWN_EVENT)

    def __call__(self, event_id: int, timeout: float | None = None) -> str:
        return super()._call_cmd(
            cmd_args=[(HdcDataType.UINT8, event_id), ],
            return_types=HdcDataType.UTF8,
            timeout=timeout
        )


class EventProxyBase:
    feature_proxy: FeatureProxyBase
    event_id: int
    most_recently_received_event_payloads: collections.deque
    payload_parser: typing.Callable[[bytes], typing.Any]

    def __init__(self,
                 feature_proxy: FeatureProxyBase,
                 event_id: int,
                 payload_parser: typing.Type[object] | None = None,
                 deque_capacity: int = 100):

        # Looks like an instance-attribute, but it's more of a class-attribute, actually. ;-)
        # Logger-name like: "hdcproto.host.proxy.MyDeviceProxy.MyFeatureProxy.MyEventProxy"
        self.logger = feature_proxy.logger.getChild(self.__class__.__name__)

        self.logger.debug(f"Initializing instance of {self.__class__.__name__} "
                          f"as proxy for EventID=0x{event_id:02X} on FeatureID=0x{feature_proxy.feature_id:02X}")

        if not is_valid_uint8(event_id):
            raise ValueError(f"event_id value of 0x{event_id:02X} is beyond valid range from 0x00 to 0xFF")

        self.event_id = event_id
        self.feature_proxy = feature_proxy

        if payload_parser is None:
            payload_parser = self.RawEventPayload  # Default payload parser for very lazy people

        self.payload_parser = payload_parser

        self.most_recently_received_event_payloads = collections.deque(maxlen=deque_capacity)

        self.event_payload_handlers = list()

        self.feature_proxy.router_feature.register_event_handler(self.event_id, self._event_message_handler)

    def _event_message_handler(self, event_message: bytes) -> None:
        """
        The one and only event handler actually registered with the RouterFeature.
        WARNING: This handler will be executed in the receiver-thread!
        """
        event_payload = self.payload_parser(event_message)
        self.most_recently_received_event_payloads.append(event_payload)
        for handler in self.event_payload_handlers:
            handler(event_payload)

    def register_event_payload_handler(self, event_payload_handler: typing.Callable[[typing.Any], None]):
        """
        Users of this proxy can register their own event-handlers.
        Event handlers should be a callable that takes the parsed event-payload object as an argument.
        Multiple handlers can be registered. They will be executed in the same order as registered.

        WARNING: Event handlers will be executed in the receiver-thread. Ensure they are fast and thread-safe!
        """
        self.event_payload_handlers.append(event_payload_handler)

    class RawEventPayload:
        """
        Default event payload parser that simply strips the message header away, keeping the raw bytes of the payload.
        It also keeps a reception timestamp.
        """

        def __init__(self, event_message: bytes):
            self.received_at = datetime.utcnow()
            self.raw_payload = event_message[3:]  # Strip 3 leading bytes: MsgID + FeatureID + EvtID


class LogEventProxy(EventProxyBase):
    logger: logging.Logger

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy,
                         event_id=EvtID.LOG,
                         payload_parser=LogEventProxy.LogEventPayload)
        # This is how HDC-logging is mapped directly into python logging:
        self.register_event_payload_handler(lambda e: self.logger.log(level=e.log_level, msg=e.log_message))

    class LogEventPayload:
        def __init__(self, event_message: bytes):
            self.received_at = datetime.utcnow()
            self.log_level = event_message[3]
            self.log_message = event_message[4:].decode(encoding="utf-8", errors="strict")

    def set_log_threshold(self, log_level: int) -> None:
        """
        This will set the same log-level on both: the HDC-LogEvent and the Python logger in this proxy.
        """
        if log_level not in [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL]:
            raise ValueError(f"A log_level of {log_level} is invalid")

        self.logger.setLevel(log_level)
        self.feature_proxy.prop_log_event_threshold.set(log_level)

    def get_log_threshold(self) -> int:
        # Note how the LogEventProxy is configured to an infinite default_freshness, thus
        # its value will always be taken from the cache.
        return self.feature_proxy.prop_log_event_threshold.get()


class StateTransitionEventProxy(EventProxyBase):
    logger: logging.Logger

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy,
                         event_id=EvtID.FEATURE_STATE_TRANSITION,
                         payload_parser=StateTransitionEventProxy.StateTransitionEventPayload)
        self.register_event_payload_handler(self.event_payload_handler)

    class StateTransitionEventPayload:
        def __init__(self, event_message: bytes):
            self.received_at = datetime.utcnow()
            self.previous_state_id = event_message[3]
            self.current_state_id = event_message[4]

    def event_payload_handler(self, event_payload: StateTransitionEventProxy.StateTransitionEventPayload):
        self.logger.info(f"%s → %s",
                         self.feature_proxy.resolve_state_name(event_payload.previous_state_id),
                         self.feature_proxy.resolve_state_name(event_payload.current_state_id))
        # Inject new state into the cache for the FeatureState property
        self.feature_proxy.prop_feature_state.update_cached_value(event_payload.current_state_id)


class PropertyProxyBase:
    feature_proxy: FeatureProxyBase
    property_id: int
    property_data_type: HdcDataType
    is_readonly: bool
    default_freshness: float
    default_timeout: float
    _cached_value: bool | int | float | str | bytes | None
    _timestamp_of_cached_value: float

    def __init__(self,
                 feature_proxy: FeatureProxyBase,
                 property_id: int,
                 property_data_type: HdcDataType,
                 is_readonly: bool,
                 default_freshness: float,
                 default_timeout: float):

        # Looks like an instance-attribute, but it's more of a class-attribute, actually. ;-)
        # Logger-name like: "hdcproto.host.proxy.MyDeviceProxy.MyFeatureProxy.MyPropertyProxy"
        self.logger = feature_proxy.logger.getChild(self.__class__.__name__)

        self.logger.debug(f"Initializing instance of {self.__class__.__name__} "
                          f"as proxy for PropertyID=0x{property_id:02X} on FeatureID=0x{feature_proxy.feature_id:02X}")

        if not is_valid_uint8(property_id):
            raise ValueError(f"property_id value of {property_id} is beyond valid range from 0x00 to 0xFF")

        self.feature_proxy = feature_proxy
        self.property_id = property_id
        self.property_data_type = property_data_type
        self.is_readonly = is_readonly

        # Keep defaults per instance, since each property may need to tweak its own sensible defaults
        self.default_freshness = default_freshness
        self.default_timeout = default_timeout

        self._cached_value = None
        self._timestamp_of_cached_value = 0.0

    def update_cached_value(self, new_value: bool | int | float | str | bytes):
        self._cached_value = new_value
        self._timestamp_of_cached_value = time.perf_counter()

    def _get(self,
             freshness: float | None = None,
             timeout: float | None = None
             ) -> bool | int | float | str | bytes:

        if freshness is None:
            freshness = self.default_freshness

        if timeout is None:
            timeout = self.default_timeout

        age_of_cached_value = time.perf_counter() - self._timestamp_of_cached_value
        if age_of_cached_value > freshness or self._cached_value is None:
            self.logger.debug(f"Getting value of PropertyID=0x{self.property_id:02X}")
            property_value = self.feature_proxy.cmd_get_property_value(property_id=self.property_id,
                                                                       property_data_type=self.property_data_type,
                                                                       timeout=timeout)
            self.logger.info(f"PropertyID=0x{self.property_id:02X} getter returns {property_value}")
            self.update_cached_value(property_value)
        else:
            self.logger.info(f"PropertyID=0x{self.property_id:02X} getter "
                             f"returns {self._cached_value} from cache "
                             f"which was updated {age_of_cached_value:.3f}s ago "
                             f"and thus is fresher than {freshness:.3f}s")

        return self._cached_value

    def _set(self,
             new_value: bool | int | float | str | bytes,
             timeout: float | None = None
             ) -> bool | int | float | str | bytes:

        self.logger.info(f"Setting PropertyID=0x{self.property_id:02X} to a value of {new_value}")

        if self.is_readonly:
            raise RuntimeError()

        if timeout is None:
            timeout = self.default_timeout

        # Note how the value returned with the reply is the actual value set on
        # the device, and it may differ from the value sent in the request!
        property_value = self.feature_proxy.cmd_set_property_value(
            property_id=self.property_id,
            property_data_type=self.property_data_type,
            new_value=new_value,
            timeout=timeout)

        self.update_cached_value(property_value)

        if property_value != new_value:
            self.logger.warning(f"Attempted to set PropertyID=0x{self.property_id:02X} "
                                f"to a value of {new_value}, but "
                                f"effectively a value of {property_value} was set.")

        self.logger.debug(f"Completed setting value of PropertyID=0x{self.property_id:02X}")

        return self._cached_value


# noinspection PyPep8Naming
class PropertyProxy_RO_BOOL(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.BOOL, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> bool:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_BOOL(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.BOOL, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> bool:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: bool, timeout: float | None = None) -> bool:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_UINT8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT8, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_UINT8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT8, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_UINT16(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT16, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_UINT16(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT16, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_UINT32(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT32, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_UINT32(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UINT32, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_INT8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT8, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_INT8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT8, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_INT16(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT16, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_INT16(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT16, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_INT32(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT32, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_INT32(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.INT32, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> int:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: int, timeout: float | None = None) -> int:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_FLOAT(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.FLOAT, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> float:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_FLOAT(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.FLOAT, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> float:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: float, timeout: float | None = None) -> float:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_DOUBLE(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.DOUBLE, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> float:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_DOUBLE(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.DOUBLE, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> float:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: float, timeout: float | None = None) -> float:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_UTF8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UTF8, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> str:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_UTF8(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.UTF8, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> str:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: str, timeout: float | None = None) -> str:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RO_BLOB(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.BLOB, is_readonly=True,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> bytes:
        return self._get(freshness=freshness, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_RW_BLOB(PropertyProxyBase):

    def __init__(self, feature_proxy: FeatureProxyBase, property_id: int,
                 default_freshness: float = 0.0,
                 default_timeout: float = DEFAULT_REPLY_TIMEOUT):
        super().__init__(feature_proxy, property_id, HdcDataType.BLOB, is_readonly=False,
                         default_freshness=default_freshness,
                         default_timeout=default_timeout)

    def get(self, freshness: float | None = None, timeout: float | None = None) -> bytes:
        return self._get(freshness=freshness, timeout=timeout)

    def set(self, new_value: bytes, timeout: float | None = None) -> bytes:
        return self._set(new_value=new_value, timeout=timeout)


# noinspection PyPep8Naming
class PropertyProxy_FeatureState(PropertyProxy_RO_UINT8):
    """
    Special handling for the FeatureState property, for which we know that we can rely on
    the cached value to be kept up to date by the FeatureStateTransitionEvent.
    This proxy will by default return the cached value and thus avoid unnecessary requests.
    Callers may explicitly force a request by calling .get(freshness=0.0)
    """

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy,
                         property_id=PropID.FEAT_STATE,
                         # Infinity, meaning that the cached value will not expire (by default)
                         default_freshness=float('inf'))

    def get_value_name(self, freshness: float | None = None, timeout: float | None = None) -> str:
        """Human-readable name of the numeric FeatureState."""
        return self.feature_proxy.resolve_state_name(self.get(freshness=freshness, timeout=timeout))


# noinspection PyPep8Naming
class PropertyProxy_LogEventThreshold(PropertyProxy_RW_UINT8):

    def __init__(self, feature_proxy: FeatureProxyBase):
        super().__init__(feature_proxy, property_id=PropID.LOG_EVT_THRESHOLD)

    def get_value_name(self, freshness: float | None = None, timeout: float | None = None) -> str:
        """Human-readable name of the numeric LogLevel-Threshold."""
        uint8_level = self.get(freshness=freshness, timeout=timeout)
        return logging.getLevelName(uint8_level)


class FeatureProxyBase:
    router_feature: hdcproto.host.router.RouterFeature
    state_names_by_id: dict[int, str]

    def __init__(self,
                 device_proxy: DeviceProxyBase,
                 feature_id: int,
                 state_names_by_id: dict[int, str] | enum.EnumMeta | None = None):
        # Looks like an instance-attribute, but it's more of a class-attribute, actually. ;-)
        # Logger-name like: "hdcproto.host.proxy.MyDeviceProxy.MyFeatureProxy"
        self.logger = device_proxy.logger.getChild(self.__class__.__name__)

        self.logger.debug(f"Initializing instance of {self.__class__.__name__} "
                          f"as proxy for FeatureID=0x{feature_id:02X}")

        if not is_valid_uint8(feature_id):
            raise ValueError(f"feature_id value of 0x{feature_id:02X} is beyond valid range from 0x00 to 0xFF")

        self.router_feature = hdcproto.host.router.RouterFeature(router=device_proxy.router,
                                                                 feature_id=feature_id)
        self.state_names_by_id = dict()

        if state_names_by_id:
            self.register_states(state_names_by_id)

        # Commands
        self.cmd_get_property_name = GetPropertyNameCommandProxy(self)
        self.cmd_get_property_type = GetPropertyTypeCommandProxy(self)
        self.cmd_get_property_readonly = GetPropertyReadonlyCommandProxy(self)
        self.cmd_get_property_value = GetPropertyValueCommandProxy(self)
        self.cmd_set_property_value = SetPropertyValueCommandProxy(self)
        self.cmd_get_property_description = GetPropertyDescriptionCommandProxy(self)
        self.cmd_get_command_name = GetCommandNameCommandProxy(self)
        self.cmd_get_command_description = GetCommandDescriptionCommandProxy(self)
        self.cmd_get_event_name = GetEventNameCommandProxy(self)
        self.cmd_get_event_description = GetEventDescriptionCommandProxy(self)

        # Events
        self.evt_log = LogEventProxy(self)
        self.evt_state_transition = StateTransitionEventProxy(self)

        # Properties (immutable)
        inf = float('inf')  # Infinity, meaning that the cached value will not expire (by default)
        self.prop_feature_name = PropertyProxy_RO_UTF8(self, property_id=PropID.FEAT_NAME,
                                                       default_freshness=inf)
        self.prop_feature_type_name = PropertyProxy_RO_UTF8(self, property_id=PropID.FEAT_TYPE_NAME,
                                                            default_freshness=inf)
        self.prop_feature_type_revision = PropertyProxy_RO_UINT8(self, property_id=PropID.FEAT_TYPE_REV,
                                                                 default_freshness=inf)
        self.prop_feature_description = PropertyProxy_RO_UTF8(self, property_id=PropID.FEAT_DESCR,
                                                              default_freshness=inf)
        self.prop_feature_tags = PropertyProxy_RO_UTF8(self, property_id=PropID.FEAT_TAGS,
                                                       default_freshness=inf)
        self.prop_available_commands = PropertyProxy_RO_BLOB(self, property_id=PropID.AVAIL_CMD,
                                                             default_freshness=inf)
        self.prop_available_events = PropertyProxy_RO_BLOB(self, property_id=PropID.AVAIL_EVT,
                                                           default_freshness=inf)
        self.prop_available_properties = PropertyProxy_RO_BLOB(self, property_id=PropID.AVAIL_PROP,
                                                               default_freshness=inf)
        # Properties (mutable)
        self.prop_feature_state = PropertyProxy_FeatureState(self)
        self.prop_log_event_threshold = PropertyProxy_LogEventThreshold(self)

    @property
    def feature_id(self) -> int:
        return self.router_feature.feature_id

    def register_state(self, state_id: int, state_name: str):
        if not is_valid_uint8(state_id):
            raise ValueError(f"state_id value of 0x{state_id:02X} is beyond valid range from 0x00 to 0xFF")

        if state_id in self.state_names_by_id:
            self.logger.warning(f"Re-registering state_id=0x{state_id:02X} as '{state_name}'. "
                                f"Previously registered as '{self.state_names_by_id[state_id]}'")
        self.state_names_by_id[state_id] = state_name

    def register_states(self, state_names_by_id: dict[int, str] | enum.EnumMeta):
        if isinstance(state_names_by_id, enum.EnumMeta):
            state_names_by_id = {e.value: e.name for e in state_names_by_id}
        elif isinstance(state_names_by_id, dict):
            pass
        else:
            raise TypeError()

        for state_id, state_name in state_names_by_id.items():
            assert isinstance(state_id, int)
            assert isinstance(state_name, str)
            self.register_state(state_id=state_id, state_name=state_name)

    def resolve_state_name(self, state_id: int) -> str:
        if not self.state_names_by_id:
            self.logger.warning(f"Can't resolve name of FeatureStateID 0x{state_id:02X}, because "
                                f"no states were registered with this proxy.")

        if not state_id in self.state_names_by_id:
            self.logger.warning(f"Can't resolve name of FeatureStateID 0x{state_id:02X}, because "
                                f"no name was registered for this ID.")
            return f"0x{state_id:02X}"  # Use hexadecimal representation as a fallback

        return self.state_names_by_id[state_id]

    def await_state(self,
                    exits: int | typing.Iterable[int] | None = None,
                    enters: int | typing.Iterable[int] | None = None,
                    timeout: float = 5.0,
                    polling_period: float = 0.1):
        """
        Utility method to block execution until the expected FeatureState(s) is/are exited and/or entered.

        Waits until the FeatureState does not match any of the states listed in :param exits:
        Then waits until the FeatureState matches any of the states listed in :param enters:
        Will raise a TimeoutError if the whole process takes longer than :param timeout: seconds.
        """
        if exits is None and enters is None:
            raise ValueError()

        if isinstance(exits, int):
            exits = [exits, ]

        if isinstance(enters, int):
            enters = [enters, ]

        deadline = time.perf_counter() + timeout

        if exits is not None:
            if any(not is_valid_uint8(v) for v in exits):
                raise ValueError("State ID values must be in range 0x00 to 0xFF")

            while self.prop_feature_state.get() in exits:
                if time.perf_counter() > deadline:
                    raise TimeoutError("Timeout before exiting state")
                time.sleep(polling_period)

        if enters is not None:
            if any(not is_valid_uint8(v) for v in enters):
                raise ValueError("State ID values must be in range 0x00 to 0xFF")

            while self.prop_feature_state.get() not in enters:
                if time.perf_counter() > deadline:
                    raise TimeoutError("Timeout before entering state")
                time.sleep(polling_period)


class CoreFeatureProxyBase(FeatureProxyBase):
    def __init__(self, device_proxy: DeviceProxyBase):
        super().__init__(device_proxy=device_proxy, feature_id=FeatureID.CORE)

        # Mandatory properties of a Core feature as required by HDC-spec
        self.prop_available_features = PropertyProxy_RO_BLOB(self, PropID.AVAIL_FEAT)
        self.prop_max_req_msg_size = PropertyProxy_RO_UINT16(self, PropID.MAX_REQ_MSG_SIZE)

        # HDC-spec does not require any mandatory commands nor events for the Core feature, other than what's
        # already mandatory for any feature, which has been inherited from the FeatureProxyBase base class.


class DeviceProxyBase:
    router: hdcproto.host.router.MessageRouter
    core: CoreFeatureProxyBase

    def __init__(self, connection_url: str):
        # Looks like an instance-attribute, but it's more of a class-attribute, actually. ;-)
        # Logger-name like: "hdcproto.host.proxy.MyDeviceProxy"
        self.logger = logger.getChild(self.__class__.__name__)

        self.logger.debug(f"Initializing instance of {self.__class__.__name__} "
                          f"as proxy for device connected at '{connection_url}'")

        serial_transport = hdcproto.transport.serialport.SerialTransport(serial_url=connection_url)
        self.router = hdcproto.host.router.MessageRouter(transport=serial_transport)

        # The following may be needed for introspection when using bare DeviceProxyBase objects.
        # Subclasses will typically override it with a more specific core-feature proxy.
        self.core = CoreFeatureProxyBase(self)

    def get_hdc_version_string(self, timeout: float = 0.2) -> str:
        """Returns the raw string, without attempting to validate nor parsing it."""
        request_message = bytes([MessageTypeID.HDC_VERSION])
        reply_message = self.router.send_request_and_get_reply(request_message, timeout)
        reply_payload = reply_message[1:]
        reply_string = reply_payload.decode(encoding="utf-8", errors="strict")
        return reply_string

    def get_hdc_version(self, timeout: float = 0.2) -> semver.VersionInfo:
        """Validates device's reply, parses version and returns it as a semver.VersionInfo object."""
        reply_string = self.get_hdc_version_string(timeout=timeout)
        expected_prefix = "HDC "
        if reply_string.startswith(expected_prefix):
            version_string = reply_string[len(expected_prefix):]
            if semver.VersionInfo.isvalid(version_string):
                return semver.VersionInfo.parse(version_string)

        raise HdcError(f"Don't know how to handle HDC-spec '{reply_string}'")

    def get_echo(self, echo_payload: bytes, timeout: float = 0.2) -> bytes:
        request_message = bytearray()
        request_message.append(MessageTypeID.ECHO)
        request_message.extend(echo_payload)
        reply_message = self.router.send_request_and_get_reply(request_message, timeout)
        reply_payload = reply_message[1:]  # Skip MessageTypeID prefix
        return reply_payload


class HdcReplyError(HdcError):
    reply_message: bytes

    def __init__(self, error_name: str, reply_message: bytes):
        self.reply_message = reply_message
        super().__init__(error_name)

    @property
    def error_code(self) -> int:
        return self.reply_message[3]

    @property
    def error_description(self) -> str:
        return self.reply_message[4:].decode(encoding="utf-8", errors="strict")



from __future__ import annotations

import enum
import struct
import typing


class HdcError(Exception):
    error_name: str

    def __init__(self, error_description: str):
        self.error_name = error_description


@enum.unique
class MessageTypeID(enum.IntEnum):
    """As defined by HDC-spec"""
    HDC_VERSION = 0xF0
    ECHO = 0xF1
    COMMAND = 0xF2
    EVENT = 0xF3


@enum.unique
class FeatureID(enum.IntEnum):
    """Reserved ID of the only mandatory HDC-feature required by HDC-spec: the Core feature."""
    CORE = 0x00


@enum.unique
class CmdID(enum.IntEnum):
    """Reserved IDs of mandatory Commands required by HDC-spec"""
    GET_PROP_NAME = 0xF0
    GET_PROP_TYPE = 0xF1
    GET_PROP_RO = 0xF2
    GET_PROP_VALUE = 0xF3
    SET_PROP_VALUE = 0xF4
    GET_PROP_DESCR = 0xF5
    GET_CMD_NAME = 0xF6
    GET_CMD_DESCR = 0xF7
    GET_EVT_NAME = 0xF8
    GET_EVT_DESCR = 0xF9


@enum.unique
class ReplyErrorCode(enum.IntEnum):
    """Reserved IDs and names of error codes used in replies to Commands as defined by HDC-spec"""
    NO_ERROR = 0x00
    UNKNOWN_FEATURE = 0xF0
    UNKNOWN_COMMAND = 0xF1
    UNKNOWN_PROPERTY = 0xF2
    UNKNOWN_EVENT = 0xF3
    INCORRECT_COMMAND_ARGUMENTS = 0xF4
    COMMAND_NOT_ALLOWED_NOW = 0xF5
    COMMAND_FAILED = 0xF6
    INVALID_PROPERTY_VALUE = 0xF7
    PROPERTY_IS_READ_ONLY = 0xF8

    def __str__(self):
        if self == ReplyErrorCode.NO_ERROR:
            return "No error"
        elif self == ReplyErrorCode.UNKNOWN_FEATURE:
            return "Unknown feature"
        elif self == ReplyErrorCode.UNKNOWN_COMMAND:
            return "Unknown command"
        elif self == ReplyErrorCode.INCORRECT_COMMAND_ARGUMENTS:
            return "Incorrect command arguments"
        elif self == ReplyErrorCode.COMMAND_NOT_ALLOWED_NOW:
            return "Command not allowed now"
        elif self == ReplyErrorCode.COMMAND_FAILED:
            return "Command failed"
        elif self == ReplyErrorCode.UNKNOWN_PROPERTY:
            return "Unknown property"
        elif self == ReplyErrorCode.INVALID_PROPERTY_VALUE:
            return "Invalid property value"
        elif self == ReplyErrorCode.PROPERTY_IS_READ_ONLY:
            return "Property is read-only"
        elif self == ReplyErrorCode.UNKNOWN_EVENT:
            return "Unknown event"


@enum.unique
class EvtID(enum.IntEnum):
    """Reserved IDs of mandatory Events required by HDC-spec"""
    LOG = 0xF0
    FEATURE_STATE_TRANSITION = 0xF1


@enum.unique
class HdcDataType(enum.IntEnum):
    """
    All IDs of data-types defined by HDC-spec.
    Mainly used to define data-type of Properties, but also used
    to parse arguments and return values of Commands.
    Also implements serialization and de-serialization from raw bytes.

    The ID value of each HdcDataType can be interpreted as follows:

    Upper Nibble: Kind of HdcDataType
          0x0_ --> Unsigned integer number
          0x1_ --> Signed integer number
          0x2_ --> Floating point number
          0xB_ --> Binary data
                   (Either variable size 0xBF, or boolean 0xB0)
          0xF_ --> UTF-8 encoded string
                   (Always variable size: 0xFF)

    Lower Nibble: Size of HdcDataType, given in number of bytes
                  i.e. 0x14 --> INT32, whose size is 4 bytes
                  (Exception to the rule: 0x_F denotes a variable size HdcDataType)
                  (Exception to the rule: 0xB0 --> BOOL, whose size is 1 bytes)
    """

    UINT8 = 0x01
    UINT16 = 0x02
    UINT32 = 0x04
    INT8 = 0x11
    INT16 = 0x12
    INT32 = 0x14
    FLOAT = 0x24
    DOUBLE = 0x28
    BOOL = 0xB0
    BLOB = 0xBF
    UTF8 = 0xFF

    def struct_format(self) -> str | None:
        if self == HdcDataType.BOOL:
            return "?"
        if self == HdcDataType.UINT8:
            return "B"
        if self == HdcDataType.UINT16:
            return "<H"
        if self == HdcDataType.UINT32:
            return "<I"
        if self == HdcDataType.INT8:
            return "<b"
        if self == HdcDataType.INT16:
            return "<h"
        if self == HdcDataType.INT32:
            return "<i"
        if self == HdcDataType.FLOAT:
            return "<f"
        if self == HdcDataType.DOUBLE:
            return "<d"
        if self == HdcDataType.BLOB:
            return None  # Meaning: Variable size
        if self == HdcDataType.UTF8:
            return None  # Meaning: Variable size

    def size(self) -> int | None:
        """
        Number of bytes of the given data type.
        Returns None for variable size types, e.g. UTF8 or BLOB
        """
        fmt = self.struct_format()
        if fmt is None:
            return None

        return struct.calcsize(fmt)

    def value_to_bytes(self, value: int | float | str | bytes) -> bytes:

        if isinstance(value, str):
            if self == HdcDataType.UTF8:
                return value.encode(encoding="utf-8", errors="strict")
            raise HdcError(f"Improper target data type {self.name} for a str value")

        if isinstance(value, bytes):
            if self == HdcDataType.BLOB:
                return value
            raise HdcError(f"Improper target data type {self.name} for a bytes value")

        fmt = self.struct_format()

        if fmt is None:
            raise HdcError(f"Don't know how to convert into {self.name}")

        if isinstance(value, bool):
            if self == HdcDataType.BOOL:
                return struct.pack(fmt, value)
            else:
                raise HdcError(f"Vale of type {value.__class__} is unsuitable "
                               f"for a property of type {self.name}")

        if isinstance(value, int):
            if self in (HdcDataType.UINT8,
                        HdcDataType.UINT16,
                        HdcDataType.UINT32,
                        HdcDataType.INT8,
                        HdcDataType.INT16,
                        HdcDataType.INT32):
                return struct.pack(fmt, value)
            else:
                raise HdcError(f"Vale of type {value.__class__} is unsuitable "
                               f"for a property of type {self.name}")

        if isinstance(value, float):
            if self in (HdcDataType.FLOAT,
                        HdcDataType.DOUBLE):
                return struct.pack(fmt, value)
            else:
                raise HdcError(f"Vale of type {value.__class__} is unsuitable "
                               f"for a property of type {self.name}")

        raise HdcError(f"Don't know how to convert value of type {value.__class__} "
                       f"into property of type {self.name}")

    def bytes_to_value(self, value_as_bytes: bytes) -> int | float | str | bytes:

        if self == HdcDataType.UTF8:
            return value_as_bytes.decode(encoding="utf-8", errors="strict")

        if self == HdcDataType.BLOB:
            return value_as_bytes

        fmt = self.struct_format()

        if fmt is None:
            raise HdcError(f"Don't know how to convert bytes of property type {self.name} "
                           f"into a python type")

        # Sanity check data size
        expected_size = self.size()
        if len(value_as_bytes) != expected_size:
            raise HdcError(
                f"Mismatch of data size. "
                f"Expected {expected_size} bytes, "
                f"but attempted to convert {len(value_as_bytes)}")

        return struct.unpack(fmt, value_as_bytes)[0]

    @staticmethod
    def parse_payload(raw_payload: bytes,
                      expected_data_types: HdcDataType | list[HdcDataType] | None
                      ) -> typing.Any:

        if expected_data_types == [None]:  # Being tolerant with weird ways of saying 'void'
            expected_data_types = None

        if not expected_data_types:
            if len(raw_payload) > 0:
                raise ValueError("Payload was expected to be empty, but it isn't.")
            return None

        return_as_list = True  # unless...
        if isinstance(expected_data_types, HdcDataType):
            return_as_list = False  # Reminder about caller not expecting a list, but a single value, instead.
            expected_data_types = [expected_data_types, ]  # Just for it to work in the for-loops below

        if any(not isinstance(t, HdcDataType) for t in expected_data_types):
            raise TypeError("Only knows how to parse for HdcDataType. (Build-in python types are not supported)")

        return_values = list()
        for idx, return_data_type in enumerate(expected_data_types):
            size = return_data_type.size()
            if size is None:
                # A size of None means it is variable length, which
                # is only allowed as last of the expected values!
                if idx != len(expected_data_types) - 1:
                    raise ValueError("Variable size values (UTF8, BLOB) are only allowed as last item")
                else:
                    size = len(raw_payload)  # Assume that the remainder of the payload is the actual value size
            if size > len(raw_payload):
                raise ValueError("Payload is shorter than expected.")
            return_value_as_bytes = raw_payload[:size]
            return_value = return_data_type.bytes_to_value(return_value_as_bytes)
            return_values.append(return_value)
            raw_payload = raw_payload[size:]

        if len(raw_payload) > 0:
            raise ValueError("Payload is longer than expected.")

        if not return_as_list:
            assert len(expected_data_types) == 1
            return return_values[0]  # Return first item, without enclosing it in a list.

        return return_values

    @staticmethod
    def parse_reply_msg(reply_message: bytes,
                        expected_data_types: HdcDataType | list[HdcDataType] | None) -> typing.Any:
        raw_payload = reply_message[4:]  # Strip 4 leading bytes: MsgID + FeatureID + EvtID + ReplyErrorCode
        return HdcDataType.parse_payload(raw_payload=raw_payload, expected_data_types=expected_data_types)

    @staticmethod
    def parse_event_msg(event_message: bytes,
                        expected_data_types: HdcDataType | list[HdcDataType] | None) -> typing.Any:
        raw_payload = event_message[3:]  # Strip 3 leading bytes: MsgID + FeatureID + EvtID
        return HdcDataType.parse_payload(raw_payload=raw_payload, expected_data_types=expected_data_types)


def is_valid_uint8(value_to_check: int) -> bool:
    """Utility method to validate whether a given int is a valid UINT8 value in an efficient and readable manner."""
    if not isinstance(value_to_check, int):
        return False  # Be stoic about it. Let caller take care to raise a more specific Exception.
    return 0x00 <= value_to_check <= 0xFF


@enum.unique
class PropID(enum.IntEnum):
    """Reserved IDs of mandatory Properties required by HDC-spec"""
    FEAT_NAME = 0xF0
    FEAT_TYPE_NAME = 0xF1
    FEAT_TYPE_REV = 0xF2
    FEAT_DESCR = 0xF3
    FEAT_TAGS = 0xF4
    AVAIL_CMD = 0xF5
    AVAIL_EVT = 0xF6
    AVAIL_PROP = 0xF7
    FEAT_STATE = 0xF8
    LOG_EVT_THRESHOLD = 0xF9

    AVAIL_FEAT = 0xFA
    """List of available features on a device (Only mandatory for the Core feature)"""

    MAX_REQ_MSG_SIZE = 0xFB
    """Largest request-message a device can cope with (Only mandatory for the Core feature)"""



from __future__ import annotations

import typing
import unittest

from hdcproto.transport.serialport import SerialTransport


class TestableSerialTransport(SerialTransport):
    received_messages: list[bytes]
    connection_loss_exception: Exception | None

    def __init__(self):
        super().__init__(serial_url='loop://')
        self.received_messages = list()
        self.connection_loss_exception = None
        self.message_received_handler = self.handle_message
        self.connection_lost_handler = self.handle_connection_loss

    def handle_message(self, message: bytes):
        self.received_messages.append(message)

    def handle_connection_loss(self, e: Exception):
        self.connection_loss_exception = e


class TestMessageRoundTrip(unittest.TestCase):

    def assertEqualAfterRoundTrip(self, messages: typing.List[bytes]):
        with TestableSerialTransport() as transport:
            for message in messages:
                transport.send_message(message)

        self.assertSequenceEqual(messages, transport.received_messages)

    def test_one_empty_message(self):
        self.assertEqualAfterRoundTrip([b''])

    def test_two_empty_messages(self):
        self.assertEqualAfterRoundTrip([b'', b''])

    def test_one_1byte_message(self):
        self.assertEqualAfterRoundTrip([b'\x00'])

    def test_two_1byte_messages(self):
        self.assertEqualAfterRoundTrip([b'\x00', b'\x00'])

    def test_one_254byte_message(self):
        self.assertEqualAfterRoundTrip([bytes(range(254))])

    def test_two_254byte_messages(self):
        self.assertEqualAfterRoundTrip([bytes(range(254)), bytes(range(254))])

    def test_one_255byte_message(self):
        self.assertEqualAfterRoundTrip([bytes(range(255))])

    def test_two_255byte_messages(self):
        self.assertEqualAfterRoundTrip([bytes(range(255)), bytes(range(255))])

    def test_one_256byte_message(self):
        self.assertEqualAfterRoundTrip([bytes(range(256))])

    def test_two_256byte_messages(self):
        self.assertEqualAfterRoundTrip([bytes(range(256)), bytes(range(256))])

    def test_one_2000byte_message(self):
        self.assertEqualAfterRoundTrip([bytes(i % 255 for i in range(2000))])

    def test_two_2000byte_messages(self):
        self.assertEqualAfterRoundTrip([bytes(i % 255 for i in range(2000)), bytes(i % 255 for i in range(2000))])

    def test_hundred_60byte_messages(self):
        self.assertEqualAfterRoundTrip([bytes(range(60))] * 100)


if __name__ == '__main__':
    unittest.main()



"""
Proxy classes to communicate with i.e. a NUCLEO prototype board running any of the Demo_Minimal STM32 firmware examples.
"""
import enum
from datetime import datetime

from hdcproto.common import HdcDataType
from hdcproto.host.proxy import (DeviceProxyBase, CoreFeatureProxyBase, VoidWithoutArgsCommandProxy, EventProxyBase,
                                 PropertyProxy_RO_UINT32, PropertyProxy_RO_BLOB, PropertyProxy_RW_UINT8)


class MinimalCore:

    def __init__(self, device_proxy: DeviceProxyBase):

        # We could inherit from CoreFeatureProxyBase, but we choose composition, instead, because
        # it allows us to separate more cleanly our custom proxies from those defined in DeviceProxyBase.
        # This is for example useful to keep the autocompletion list short and readable.
        # https://en.wikipedia.org/wiki/Composition_over_inheritance
        self.hdc = CoreFeatureProxyBase(device_proxy=device_proxy)

        self.hdc.register_states(MinimalCore.FeatureStateEnum)

        # Commands
        # ToDo: A reset command handler is responsible to evict any cached values/states on the proxies!
        self.cmd_reset = VoidWithoutArgsCommandProxy(self.hdc, command_id=0xC1, default_timeout=1.23)

        # Events
        self.evt_button = EventProxyBase(self.hdc, event_id=0x01, payload_parser=self.ButtonEventPayload)

        # Properties
        self.prop_microcontroller_devid = PropertyProxy_RO_UINT32(self.hdc, property_id=0x010)
        self.prop_microcontroller_revid = PropertyProxy_RO_UINT32(self.hdc, property_id=0x11)
        self.prop_microcontroller_uid = PropertyProxy_RO_BLOB(self.hdc, property_id=0x12)
        self.prop_led_blinking_rate = PropertyProxy_RW_UINT8(self.hdc, property_id=0x13)

    class FeatureStateEnum(enum.IntEnum):
        """Used by FeatureProxyBase.resolve_state_name() to resolve names of states of the MinimalCore feature."""
        OFF = 0
        INITIALIZING = 1
        READY = 2
        ERROR = 0xFF

    class ButtonEventPayload:
        """Used by the self.evt_button proxy to parse raw event messages into custom event payload objects."""

        def __init__(self, event_message: bytes):
            # Timestamp that might be handy when processing this event in a delayed manner
            self.received_at = datetime.utcnow()

            # In simple cases like this and knowing some internals of the HDC-spec, it might be better to do this:
            #     self.button_id = event_message[3]
            #     self.button_state = event_message[4]

            # Otherwise, it might be better to do this:
            self.button_id, self.button_state = HdcDataType.parse_event_msg(
                event_message=event_message,
                expected_data_types=[HdcDataType.UINT8, HdcDataType.UINT8]
            )


class MinimalDevice(DeviceProxyBase):
    # Note how this class *inherits* from DeviceProxyBase, but it could also choose to use it for "composition"
    core: MinimalCore

    def __init__(self, connection_url: str):
        super().__init__(connection_url=connection_url)

        # Demo_Minimal only implements the mandatory Core feature
        self.core = MinimalCore(self)  # Override base Core-Feature proxy with a more specific one.



"""
Showcases the resetting of features.

This script has been used to investigate the resetting behavior.
"""
import logging
import time

from minimal_proxy import MinimalDevice


def showcase_reset():
    #########################################################
    # This example uses python logging to explain the
    # demonstration, but also show internals of the HDC-host
    demo_logger = logging.getLogger('showcase_reset')
    demo_logger.setLevel(logging.INFO)
    #
    hdc_root_logger = logging.getLogger()
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter('%(asctime)s.%(msecs)03d - %(levelname)7s - %(name)s - %(message)s',
                                               datefmt='%M:%S'))
    hdc_root_logger.addHandler(log_handler)

    # You can tweak the following log-levels to tune verbosity of HDC internals:
    logging.getLogger("hdcproto.transport.packetizer").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.transport.serialport").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.host.router").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.host.proxy").setLevel(logging.DEBUG)

    #################################################
    # Connect to HDC-device at a specific serial port
    dev = MinimalDevice(connection_url="COM10")
    dev.router.connect()
    demo_logger.info(f"Device reports to be compliant with: '{dev.get_hdc_version_string()}'")

    demo_logger.info("____________________________________")
    demo_logger.info("Resetting the Core-feature...")
    dev.core.cmd_reset()
    time.sleep(1)

    demo_logger.info("____________________________________")
    demo_logger.info("Resetting the Core-feature, again...")
    dev.core.cmd_reset()
    time.sleep(1)

    dev.router.close()


if __name__ == '__main__':
    showcase_reset()



from __future__ import annotations

import logging
import threading
import time
import typing

import serial

from hdcproto.transport.base import TransportBase
from hdcproto.transport.packetizer import Packetizer


logger = logging.getLogger(__name__)  # Logger-name: "hdcproto.transport.serialport"


class SerialTransport(TransportBase):
    """
    Sends and receives messages via a serial port (typically a Virtual Com Port of a USB-CDC connection).
    Internally it uses the HDC-packetizer to allow transmission of HDC-messages, because serial-communication
    is only able to transmit raw streams of bytes.

    WARNING: Message handler callbacks will be called directly from a dedicated data-receiver-thread!
    """
    serial_url: str
    serial_port: serial.Serial | None
    receiver_thread: threading.Thread | None
    keep_thread_alive: bool
    packetizer: Packetizer
    port_access_lock = threading.Lock

    TIMEOUT_READ = 0.5  # Period of time after which we can be sure that an incoming data burst is completed

    def __init__(self,
                 serial_url: str,
                 message_received_handler: typing.Callable[[bytes], None] | None = None,
                 connection_lost_handler: typing.Callable[[Exception], None] | None = None):
        super().__init__(message_received_handler=message_received_handler,
                         connection_lost_handler=connection_lost_handler)
        self.serial_url = serial_url
        self.serial_port = None  # Will be initialized on connection
        self.receiver_thread = None  # Will be initialized on connection
        self.keep_thread_alive = True
        self.packetizer = Packetizer()
        self.port_access_lock = threading.Lock()

    def connect(self):
        if not self.message_received_handler or not self.connection_lost_handler:
            raise RuntimeError("Must assign message_received_handler and connection_lost_handler before connecting")
        logger.info(f"Connecting to {self.serial_url}")
        self.serial_port = serial.serial_for_url(self.serial_url, timeout=self.TIMEOUT_READ, baudrate=115200)
        self.receiver_thread = threading.Thread(target=self.receiver_thread_loop,
                                                kwargs={'transport': self},
                                                daemon=True)
        self.keep_thread_alive = True
        self.receiver_thread.start()

    def send_message(self, message: bytes):
        with self.port_access_lock:
            for packet in self.packetizer.pack_message(message):
                self.serial_port.write(packet)  # ToDo: Maybe we should concatenate all packets and just call this once?

    def flush(self):
        with self.port_access_lock:
            while self.serial_port.out_waiting or self.serial_port.in_waiting:
                time.sleep(0.001)

    def close(self):
        """
        Stops the receiver-thread and close the serial port.
        Does so immediately without caring for any ongoing communication
        """
        # use the lock to let other threads finish writing
        with self.port_access_lock:
            # first stop receiver-thread, so that closing can be done on idle port
            self.keep_thread_alive = False
            if hasattr(self.serial_port, 'cancel_read'):
                self.serial_port.cancel_read()
            self.receiver_thread.join(2 * SerialTransport.TIMEOUT_READ)
            self.receiver_thread = None

            # now it's safe to close the port
            self.serial_port.close()
            self.serial_port = None

    # - -  context manager
    def __enter__(self) -> SerialTransport:
        """Enter context handler. May raise RuntimeError in case the connection could not be created."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Leave context handler"""
        self.flush()
        self.close()

    def __str__(self):
        return f"{self.__class__.__name__}('{self.serial_url}')"

    # - -  Receiver-Thread
    def receiver_thread_loop(self, transport: SerialTransport) -> None:
        """
        This will be executed in a dedicated Thread, to ensure that any received messages are being handled as
        soon as they are being received.
        """
        logger.info(f"Started receiver-thread with native id: {threading.get_native_id()}")

        if transport.serial_port is None and transport.connection_lost_handler:
            transport.connection_lost_handler(Exception("Failed to start receiver-thread, because port is not open"))
            return

        packetizer = Packetizer()

        while transport.keep_thread_alive and self.serial_port.is_open:
            try:
                # Read all bytes received or wait for one byte (blocking, timing out after TIMEOUT_READ)
                # Empty data means we reached the timeout without any new bytes
                data = transport.serial_port.read(transport.serial_port.in_waiting or 1)
            except serial.SerialException as e:
                logger.exception("Serial connection has failed.")
                # probably some I/O problem such as disconnected USB serial adapters -> exit
                if transport.connection_lost_handler:
                    logger.info("About to call the connection_lost_handler.")
                    transport.connection_lost_handler(e)
                else:
                    logger.info("Not calling any connection_lost_handler.")
                return
            else:

                # No need for DEBUG logging here, because it would be redundant with what the packetizer logs.

                packetizer.data_received(data)
                received_messages = packetizer.get_received_messages()

                try:  # Catch any exception thrown by user-code that processes the received messages
                    for message in received_messages:
                        transport.message_received_handler(message)
                except Exception as e:
                    # ToDo: Should we log and ignore or disconnect with an exception?
                    if transport.connection_lost_handler:
                        logger.info("About to call the connection_lost_handler.")
                        transport.connection_lost_handler(e)
                    else:
                        logger.info("Not calling any connection_lost_handler.")
                    return


########################
# Showcase this module

def showcase_serial_transport():
    def handle_message(message: bytes):
        print(f'Received {len(message):3d} byte message: [{message.hex(sep=",")}]')

    def handle_lost_connection(exception):
        print(f'Lost connection, because: {exception}')

    with SerialTransport(serial_url="loop://",
                         message_received_handler=handle_message,
                         connection_lost_handler=handle_lost_connection) as transport:
        transport.send_message(bytes())  # Empty packet.
        transport.send_message(bytes(range(1)))
        transport.send_message(bytes(range(10)))
        transport.send_message(bytes(range(254)))
        transport.send_message(bytes(range(255)))  # Multi-packet message of 255 bytes.
        transport.send_message(bytes(i % 255 for i in range(400)))  # Multi-packet message of 400 bytes.
        transport.send_message(bytes(i % 255 for i in range(510)))  # Multi-packet message of 510 bytes.
        transport.flush()


if __name__ == '__main__':
    showcase_serial_transport()



import unittest

from hdcproto.host.proxy import (DeviceProxyBase, FeatureProxyBase, VoidWithoutArgsCommandProxy,
                                 PropertyProxy_RW_INT32, EventProxyBase)
from hdcproto.common import ReplyErrorCode


class TestableDeviceProxy(DeviceProxyBase):
    pass


class TestReplyErrorCodeRegistration(unittest.TestCase):
    def setUp(self) -> None:
        my_device = TestableDeviceProxy(connection_url="loop:")
        self.some_command_proxy = my_device.core.cmd_get_property_value

    def test_predefined_replyerrorcode_which_is_not_yet_registered(self):
        replyerrorcode_not_yet_registered = ReplyErrorCode.UNKNOWN_EVENT
        try:
            self.some_command_proxy.register_error(replyerrorcode_not_yet_registered)
        except ValueError:
            self.fail("Failed to register predefined ReplyErrorCode!")

    def test_predefined_replyerrorcode_which_is_already_registered(self):
        replyerrorcode_already_registered = ReplyErrorCode.UNKNOWN_PROPERTY
        with self.assertRaises(ValueError):
            self.some_command_proxy.register_error(replyerrorcode_already_registered)

    def test_custom_replyerrorcode_which_is_available(self):
        replyerrorcode_not_predefined_by_hdc = 0xDD
        try:
            self.some_command_proxy.register_error(replyerrorcode_not_predefined_by_hdc)
        except ValueError:
            self.fail("Failed to register custom ReplyErrorCode!")

    def test_custom_replyerrorcode_which_is_not_available(self):
        replyerrorcode_predefined_by_hdc = int(ReplyErrorCode.UNKNOWN_FEATURE)
        with self.assertRaises(ValueError):
            self.some_command_proxy.register_error(replyerrorcode_predefined_by_hdc)

    def test_custom_replyerrorcode_which_is_below_valid_range(self):
        replyerrorcode_negative = -1
        with self.assertRaises(ValueError):
            self.some_command_proxy.register_error(replyerrorcode_negative)

    def test_custom_replyerrorcode_which_is_beyond_valid_range(self):
        replyerrorcode_too_big = 256
        with self.assertRaises(ValueError):
            self.some_command_proxy.register_error(replyerrorcode_too_big)


class TestIdValidation(unittest.TestCase):
    def test_validation_of_feature_id(self):
        my_device = TestableDeviceProxy(connection_url="loop:")

        with self.assertRaises(ValueError):
            FeatureProxyBase(my_device, feature_id=-1)

        with self.assertRaises(ValueError):
            FeatureProxyBase(my_device, feature_id=256)

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            FeatureProxyBase(my_device, feature_id='42')

        # Does *not* fail with a valid feature_id
        FeatureProxyBase(my_device, feature_id=0x42)

    def test_validation_of_command_id(self):
        my_device = TestableDeviceProxy(connection_url="loop:")

        with self.assertRaises(ValueError):
            VoidWithoutArgsCommandProxy(my_device.core, command_id=-1)

        with self.assertRaises(ValueError):
            VoidWithoutArgsCommandProxy(my_device.core, command_id=256)

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            VoidWithoutArgsCommandProxy(my_device.core, command_id='42')

        # Does *not* fail with a valid command_id
        VoidWithoutArgsCommandProxy(my_device.core, command_id=0x42)

    def test_validation_of_property_id(self):
        my_device = TestableDeviceProxy(connection_url="loop:")

        with self.assertRaises(ValueError):
            PropertyProxy_RW_INT32(my_device.core, property_id=-1)

        with self.assertRaises(ValueError):
            PropertyProxy_RW_INT32(my_device.core, property_id=256)

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            PropertyProxy_RW_INT32(my_device.core, property_id='42')

        # Does *not* fail with a valid property_id
        PropertyProxy_RW_INT32(my_device.core, property_id=0x42)

    def test_validation_of_event_id(self):
        my_device = TestableDeviceProxy(connection_url="loop:")

        with self.assertRaises(ValueError):
            EventProxyBase(my_device.core, event_id=-1)

        with self.assertRaises(ValueError):
            EventProxyBase(my_device.core, event_id=256)

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            EventProxyBase(my_device.core, event_id='42')

        # Does *not* fail with a valid event_id
        EventProxyBase(my_device.core, event_id=0x42)


if __name__ == '__main__':
    unittest.main()



"""
Showcases how the (atypical) Echo-command works

This script has also been handy to debug the C implementation of the HDC-device.
"""

import logging
import time

from hdcproto.host.proxy import DeviceProxyBase


def showcase_echo():
    #########################################################
    # This example uses python logging to explain the
    # demonstration, but also show internals of the HDC-host
    demo_logger = logging.getLogger('showcase_echo')
    demo_logger.setLevel(logging.INFO)
    #
    hdc_root_logger = logging.getLogger()
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter('%(asctime)s.%(msecs)03d - %(levelname)7s - %(name)s - %(message)s',
                                               datefmt='%M:%S'))
    hdc_root_logger.addHandler(log_handler)

    # You can tweak the following log-levels to tune verbosity of HDC internals:
    logging.getLogger("hdcproto.transport.packetizer").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.transport.serialport").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.host.router").setLevel(logging.DEBUG)
    logging.getLogger("hdcproto.host.proxy").setLevel(logging.DEBUG)

    #################################################
    # Connect to HDC-device at a specific serial port
    device_proxy = DeviceProxyBase(connection_url="COM10")
    device_proxy.router.connect()

    num_requests = 100
    # Warning: HDC-device has limited buffer size for receiving a request message!
    #          Larger messages will raise reading-frame-error LogEvents!
    # Find out what the largest request message size is that this device can cope with.
    max_req_msg_size = device_proxy.core.prop_max_req_msg_size.get()
    payload_size = max_req_msg_size - 1  # Because of the MessageType byte that the Echo-request will prepend

    demo_logger.info(f"Device reports to be compliant with: '{device_proxy.get_hdc_version_string()}'")
    demo_logger.info("______________________________________________________________________")
    demo_logger.info(f"Showcasing ECHO command. Sending {num_requests} requests "
                     f"with a payload of {payload_size} bytes each:")
    timestamp_start = time.time_ns()
    sent_data = bytes(range(payload_size))
    for i in range(num_requests):
        replied_data = device_proxy.get_echo(sent_data)
        assert replied_data == sent_data
    timestamp_stop = time.time_ns()
    duration_total_ms = (timestamp_stop - timestamp_start) / 1000000.0
    demo_logger.info(f"Completed in {duration_total_ms:.1f}ms --> {duration_total_ms / num_requests:.1f} ms/request")

    device_proxy.router.close()


if __name__ == '__main__':
    showcase_echo()



from __future__ import annotations

import typing


class TransportBase:
    """
    Base class of any transport capable to sending and receiving messages to/from a device.
    """

    message_received_handler: typing.Callable[[bytes], None] | None
    connection_lost_handler: typing.Callable[[Exception], None] | None

    def __init__(self,
                 message_received_handler: typing.Callable[[bytes], None] | None = None,
                 connection_lost_handler: typing.Callable[[Exception], None] | None = None):
        self.message_received_handler = message_received_handler
        self.connection_lost_handler = connection_lost_handler

    def connect(self):
        raise NotImplementedError()

    def send_message(self, message: bytes):
        """Transmits a chunk of raw bytes to the device"""
        raise NotImplementedError()

    def flush(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __str__(self):
        raise NotImplementedError()



"""
Message router of the HDC-host implementation
"""
from __future__ import annotations

import logging
import threading
import typing

from hdcproto.common import MessageTypeID, HdcError, is_valid_uint8
from hdcproto.transport.base import TransportBase


logger = logging.getLogger(__name__)  # Logger-name: "hdcproto.host.router"


class MessageRouter:
    """
    Plugs things together:
        - receiving and sending of raw bytes over a specific transport,
        - packing and unpacking of messages to/from a stream of raw bytes
        - routing of messages received from the device
        - sending of messages to the device
    Not to be confused with the Device-Proxy classes!
    """
    features: dict[int, RouterFeature]
    transport: TransportBase
    request_reply_lock: threading.Lock
    received_reply_event: threading.Event
    last_reply_message: bytes
    strict_event_handling: bool

    def __init__(self, transport: TransportBase):
        self.features = dict()
        self.transport = transport
        transport.message_received_handler = self.handle_message
        transport.connection_lost_handler = self.handle_lost_connection
        self.request_reply_lock = threading.Lock()
        self.received_reply_event = threading.Event()
        self.last_reply_message = bytes()
        self.strict_event_handling = False

    def connect(self):
        logger.info(f"Connecting via {self.transport}")
        self.transport.connect()

    def close(self):
        logger.info(f"Closing connection via {self.transport}")
        self.transport.close()

    def handle_lost_connection(self):
        logger.error(f"Lost connection via {self.transport}")
        raise NotImplementedError()

    def handle_message(self, message: bytes):
        """
        Executed within the ReceiverThread!
        Do not use this for message processing!
        This is just about receiving and signalling it to the main thread, where the actual processing will happen!
        """
        if len(message) == 0:
            return  # Ignore empty messages, for now!

        msg_type_id = message[0]

        if msg_type_id in [MessageTypeID.HDC_VERSION, MessageTypeID.ECHO, MessageTypeID.COMMAND]:
            self.handle_requested_reply(message)
        elif msg_type_id == MessageTypeID.EVENT:
            self.handle_event(message)
        else:
            # ToDo: Should we fail silently, instead, to be forward-compatible with future versions of HDC-spec?
            raise HdcError(f"Don't know how to handle MessageTypeID=0x{msg_type_id:02X}")

    def handle_requested_reply(self, message: bytes):
        """
        Handler for any kind reply that has been explicitly requested,
           i.e.: VersionMessages, EchoMessages, CommandMessages

        Note how replies to Command-requests do not need to look up neither FeatureID nor CommandID, because
        whatever command issued the request is currently blocking and awaiting the received_reply_event.
        """
        # Note how we do

        if not self.request_reply_lock.locked():
            # If the lock is not currently taken, then no-one is awaiting this reply (anymore)!
            # i.e. requester has meanwhile timed out
            # i.e. device sent a reply that we did not request (e.g. reconnecting to an active device)
            return

        self.last_reply_message = message
        self.received_reply_event.set()

    def handle_event(self, message: bytes) -> None:
        feature_id = message[1]
        if feature_id not in self.features:
            event_id = message[2]
            if self.strict_event_handling:
                raise HdcError(f"EventID=0x{event_id:02X} raised by unknown FeatureID=0x{feature_id:02X}")
            else:
                logger.debug(f"Ignoring EventID=0x{event_id:02X} raised by unknown FeatureID=0x{feature_id:02X}")
                return

        self.features[feature_id].handle_event(message)

    def send_request_and_get_reply(self, request_message: bytes, timeout: float) -> bytes:

        # ToDo: We might need to make the following blocking=True with a timeout if the caller is multi-threaded.
        if not self.request_reply_lock.acquire(blocking=False):
            raise HdcError("Mustn't send a request if no reply was yet received for a preceding request")
        try:
            if self.received_reply_event.is_set():
                raise HdcError("Did not expect the received_reply_event to be signaled before sending a request")

            self.transport.send_message(request_message)

            if not self.received_reply_event.wait(timeout=timeout):
                raise TimeoutError(f'Did not receive any reply after {timeout} seconds.')

            self.received_reply_event.clear()

            return self.last_reply_message

        finally:
            self.request_reply_lock.release()


class RouterFeature:
    """
    This class is about everything the MessageRouter needs to know about a feature of a device.
    Not to be confused with the Feature-Proxy classes, which are meant as the actual API.
    """
    feature_id: int
    router: MessageRouter
    event_handlers: dict[int, typing.Callable[[bytes], None]]  # Handlers implemented on the proxy class

    def __init__(self, router: MessageRouter, feature_id: int):
        if not is_valid_uint8(feature_id):
            raise ValueError(f"feature_id value of {feature_id} is beyond valid range from 0x00 to 0xFF")
        self.event_handlers = dict()
        self.feature_id = feature_id
        self.router = router
        if feature_id in router.features:
            logger.warning(f"Re-registering a feature with ID {feature_id}")
        router.features[feature_id] = self

    def register_event_handler(self, event_id: int, event_handler: typing.Callable[[bytes], None]) -> None:
        if event_id in self.event_handlers:
            raise HdcError(f"Feature {self.feature_id} already has EventID {event_id}")
        self.event_handlers[event_id] = event_handler

    def handle_event(self, message: bytes):
        """
        Executed within the ReceiverThread!
        Do not use this for message processing!
        This is just about receiving and signalling it to the main thread, where the actual processing will happen!
        """
        event_id = message[2]
        if event_id not in self.event_handlers:
            if self.router.strict_event_handling:
                raise HdcError(f"Unknown EventID=0x{event_id:02X} raised by FeatureID=0x{self.feature_id:02X}")
            else:
                logger.debug(f"Ignoring EventID=0x{event_id:02X} "
                             f"raised by FeatureID=0x{self.feature_id:02X}")
                return

        event_handler = self.event_handlers[event_id]
        event_handler(message)



"""
Showcases how the capabilities of an unknown device can be explored
by means of the introspection capabilities of the HDC protocol.

https://en.wikipedia.org/wiki/Type_introspection

This script has also been handy to debug the C implementation of the HDC-device.
"""
from hdcproto.common import HdcDataType
from hdcproto.host.proxy import DeviceProxyBase, FeatureProxyBase


def is_hdc_internal(member_id: int) -> bool:
    return (member_id & 0xF0) == 0xF0


def showcase_introspection(skip_mandatory_members: bool = False):

    device_proxy = DeviceProxyBase(connection_url="COM10")
    device_proxy.router.connect()

    print(f"Device reports to be compliant with: {device_proxy.get_hdc_version_string()}")

    # Introspection: Available Features
    available_feature_ids = device_proxy.core.prop_available_features.get(timeout=6)
    print(f"Features: {available_feature_ids}")

    for featureID in available_feature_ids:
        feature_proxy = FeatureProxyBase(device_proxy=device_proxy, feature_id=featureID)
        print(f"Feature[0x{featureID:02X}]: {feature_proxy.prop_feature_name.get(timeout=600)}")

        # Introspection: Available Commands on Core feature
        available_command_ids = feature_proxy.prop_available_commands.get(timeout=600)

        for cmdID in available_command_ids:
            if skip_mandatory_members and is_hdc_internal(cmdID):
                continue
            cmd_name = feature_proxy.cmd_get_command_name(cmdID, timeout=600)
            cmd_desc = feature_proxy.cmd_get_command_description(cmdID, timeout=600).replace("\n", "\n\t\t\t")
            print(f"\tCommand[0x{cmdID:02X}]: {cmd_name} {cmd_desc}")

        # Introspection: Available Events on Core feature
        available_event_ids = feature_proxy.prop_available_events.get(timeout=600)

        for evtID in available_event_ids:
            if skip_mandatory_members and is_hdc_internal(evtID):
                continue
            evt_name = feature_proxy.cmd_get_event_name(evtID, timeout=600)
            evt_desc = feature_proxy.cmd_get_event_description(evtID, timeout=600).replace("\n", "\n\t\t\t")
            print(f"\tEvent[0x{evtID:02X}]: {evt_name} {evt_desc}")

        # Introspection: Available Properties on Core feature
        available_property_ids = feature_proxy.prop_available_properties.get(timeout=600)

        for propID in available_property_ids:
            if skip_mandatory_members and is_hdc_internal(propID):
                continue
            prop_ro = feature_proxy.cmd_get_property_readonly(propID, timeout=600)
            prop_type: HdcDataType = feature_proxy.cmd_get_property_type(propID, timeout=600)
            prop_name = feature_proxy.cmd_get_property_name(propID, timeout=600)
            prop_value = feature_proxy.cmd_get_property_value(propID, prop_type, timeout=600)
            if isinstance(prop_value, bytes):
                prop_value = ', '.join(f'0x{byte:02X}' for byte in prop_value)
                prop_value = '[' + prop_value + ']'
            prop_desc = feature_proxy.cmd_get_property_description(propID, timeout=600).replace("\n", "\n\t\t\t")
            print(f"\tProperty[0x{propID:02X}]: "
                  f"{'RO' if prop_ro else 'RW'} "
                  f"{prop_type.name} {prop_name} = {prop_value} {prop_desc}")


if __name__ == '__main__':
    showcase_introspection(skip_mandatory_members=False)



from __future__ import annotations

import typing
import unittest

from hdcproto.transport.packetizer import Packetizer


class TestChecksum(unittest.TestCase):
    def test_checksum_computation(self):
        self.assertEqual(
            Packetizer.compute_checksum(b''),
            0x00
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\x01'),
            0xFF
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\x02'),
            0xFE
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\x03'),
            0xFD
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\x01\x01\x01'),
            0xFD
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\xFF'),
            0x01
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\xFF\xFF'),
            0x02
        )

        self.assertEqual(
            Packetizer.compute_checksum(b'\xFF\xFF\xFF'),
            0x03
        )


def single_chunk(packets: list[bytes]) -> list[bytes]:
    """Concatenates all given packets into a single chunk of bytes"""
    return [b''.join(packets)]


def chunk_into_single_bytes(packets: list[bytes]) -> list[bytes]:
    """Concatenates all given packets into a single chunk of bytes and then chunks it into single byte chunks"""
    return [bytes([single_byte]) for single_byte in b''.join(packets)]


class TestMessageRoundTrip(unittest.TestCase):

    def assertEqualAfterRoundTrip(
            self,
            messages: list[bytes],
            chunker: typing.Callable[[list[bytes]], list[bytes]]):

        # messages --> packets
        packets = list()
        for message in messages:
            packets.extend(Packetizer.pack_message(message))
        self.assertGreaterEqual(len(packets), 1)

        # packets --> messages
        receiving_packetizer = Packetizer()
        for chunk in chunker(packets):
            receiving_packetizer.data_received(chunk)
        received_messages = receiving_packetizer.get_received_messages()

        self.assertSequenceEqual(
            received_messages,
            messages)

        self.assertEqual(
            len(receiving_packetizer.get_received_messages()),
            0,
            "Expected to auto-clear list of received messages!")

        self.assertEqual(
            receiving_packetizer.reading_frame_error_count,
            0)

    def test_one_empty_message(self):
        messages = [bytes()]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_empty_messages(self):
        messages = [bytes(), bytes()]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_1byte_message(self):
        messages = [bytes(range(1))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_1byte_messages(self):
        messages = [bytes(range(1)), bytes(range(1))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_254byte_message(self):
        messages = [bytes(range(254))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_254byte_messages(self):
        messages = [bytes(range(254)), bytes(range(254))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_255byte_message(self):
        messages = [bytes(range(255))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_255byte_messages(self):
        messages = [bytes(range(255)), bytes(range(255))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_256byte_message(self):
        messages = [bytes(range(256))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_256byte_messages(self):
        messages = [bytes(range(256)), bytes(range(256))]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_2550byte_message(self):
        message = bytes(i % 256 for i in range(2550))
        messages = [message]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_2550byte_message(self):
        message = bytes(i % 256 for i in range(2550))
        messages = [message, message]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_one_2551byte_message(self):
        message = bytes(i % 256 for i in range(2551))
        messages = [message]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)

    def test_two_2551byte_message(self):
        message = bytes(i % 256 for i in range(2551))
        messages = [message, message]
        self.assertEqualAfterRoundTrip(messages, single_chunk)
        self.assertEqualAfterRoundTrip(messages, chunk_into_single_bytes)


class TestReadingFrameErrors(unittest.TestCase):
    def test_bogus_first_byte_caught_by_terminator_mismatch(self):
        message = bytes(range(32))
        packetizer = Packetizer()
        packets = packetizer.pack_message(message)
        raw_data = single_chunk(packets)[0]

        # The bogus byte fools packetizer into expecting a much shorter message,
        # but it won't find any terminator where expected.
        bogus_byte = b'\x05'
        packetizer.data_received(bogus_byte + raw_data)
        self.assertEqual(len(packetizer._received_messages), 1)  # It realized something is wrong and skipped bogus byte
        unpacked_messages = packetizer.get_received_messages()
        self.assertEqual(len(packetizer._received_messages), 0)  # Auto-cleared internal message buffer
        self.assertEqual(len(unpacked_messages), 1)
        self.assertSequenceEqual(unpacked_messages[0], message)
        self.assertEqual(packetizer.reading_frame_error_count, 1)

    def test_bogus_first_byte_caught_by_checksum_mismatch(self):
        message = bytes([Packetizer.TERMINATOR] * 32)
        packetizer = Packetizer()
        packets = packetizer.pack_message(message)
        raw_data = single_chunk(packets)[0]

        # The bogus byte fools packetizer into expecting a much shorter message,
        # and it coincidentally will even find a terminator where expected, but the checksum will not match.
        bogus_byte = b'\x05'
        packetizer.data_received(bogus_byte + raw_data)
        self.assertEqual(len(packetizer._received_messages), 1)  # It realized something is wrong and skipped bogus byte
        unpacked_messages = packetizer.get_received_messages()
        self.assertEqual(len(packetizer._received_messages), 0)  # Auto-cleared internal message buffer
        self.assertEqual(len(unpacked_messages), 1)
        self.assertSequenceEqual(unpacked_messages[0], message)
        self.assertEqual(packetizer.reading_frame_error_count, 1)

    def test_bogus_first_byte_caught_by_end_of_burst(self):
        message = bytes(range(32))
        packetizer = Packetizer()
        packets = packetizer.pack_message(message)
        raw_data = single_chunk(packets)[0]

        # The bogus byte fools packetizer into expecting a much longer message
        bogus_byte = b'\x99'
        packetizer.data_received(bogus_byte + raw_data)
        self.assertEqual(len(packetizer._received_messages), 0)  # Packetizer is still hoping for more chunks to arrive
        packetizer.data_received(b'')  # End of a data burst is signalled to packetizer with an empty chunk
        self.assertEqual(len(packetizer._received_messages), 1)  # It realized something is wrong and skipped bogus byte
        unpacked_messages = packetizer.get_received_messages()
        self.assertEqual(len(packetizer._received_messages), 0)  # Auto-cleared internal message buffer
        self.assertEqual(len(unpacked_messages), 1)
        self.assertSequenceEqual(unpacked_messages[0], message)
        self.assertEqual(packetizer.reading_frame_error_count, 1)

    def test_bogus_first_5bytes(self):
        message = bytes(range(32))
        packetizer = Packetizer()
        packets = packetizer.pack_message(message)
        raw_data = single_chunk(packets)[0]
        bogus_byte = b'\x00\x00\x00\x00\x00'  # A typical burst of bogus zeroes
        packetizer.data_received(bogus_byte + raw_data)
        packetizer.data_received(b'')  # End of a data burst is signalled to packetizer with an empty chunk
        self.assertEqual(len(packetizer._received_messages), 1)  # It realized something is wrong and skipped all bogus
        unpacked_messages = packetizer.get_received_messages()
        self.assertEqual(len(packetizer._received_messages), 0)  # Auto-cleared internal message buffer
        self.assertEqual(len(unpacked_messages), 1)
        self.assertSequenceEqual(unpacked_messages[0], message)
        self.assertEqual(packetizer.reading_frame_error_count, 5)


if __name__ == '__main__':
    unittest.main()



"""
Showcases the HDC LogEvent and the Python loggers.

Note how the mandatory property LogLevelThreshold of each HDC-feature controls how much bandwidth
we may devote to LogEvents raised by each feature.

This script has also been handy to debug the C implementation of the HDC-device.
"""

import logging
import time

import hdcproto.transport.serialport
from hdcproto.host.proxy import DeviceProxyBase


def provoke_some_log_events(transport: hdcproto.transport.serialport.SerialTransport):
    transport.serial_port.write(bytes([0, 0, 0]))  # Provoke reading-frame warning
    transport.send_message(bytes([0, 0, 0]))  # Provoke unknown message type error
    time.sleep(2)  # Wait for some "heart-beat" LogEvents to happen


def showcase_logging():
    #########################################################
    # This example uses python logging to explain the
    # demonstration, but also show internals of the HDC-host
    demo_logger = logging.getLogger('showcase_logging')
    demo_logger.setLevel(logging.INFO)
    #
    hdc_root_logger = logging.getLogger()
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter('%(asctime)s.%(msecs)03d - %(levelname)7s - %(name)s - %(message)s',
                                               datefmt='%M:%S'))
    hdc_root_logger.addHandler(log_handler)

    # You can tweak the following log-levels to tune verbosity of HDC internals:
    logging.getLogger("hdcproto.transport.packetizer").setLevel(logging.WARNING)
    logging.getLogger("hdcproto.transport.serialport").setLevel(logging.WARNING)
    logging.getLogger("hdcproto.host.router").setLevel(logging.WARNING)

    ##################################################################################
    # To demonstrate the granularity of loggers in hdcproto, we'll suppress logs
    # from any other proxies, by simply not setting any level on their common parent.
    # Further below we'll only enable the LogEventProxy logger.
    logging.getLogger("hdcproto.host.proxy").setLevel(logging.NOTSET)

    #################################################
    # Connect to HDC-device at a specific serial port
    device_proxy = DeviceProxyBase(connection_url="COM10")
    device_proxy.router.connect()

    demo_logger.info("__________________________________________________________________________________________")
    demo_logger.info(f"Device reports to be compliant with: '{device_proxy.get_hdc_version_string()}'")
    demo_logger.info("__________________________________________________________________________________________")
    demo_logger.info("This demonstration will intentionally send some corrupt request to the HDC-device, for it")
    demo_logger.info("to complain via LogEvents about different degrees of severity. ")
    demo_logger.info("__________________________________________________________________________________________")
    demo_logger.info(f"Setting LogLevelThreshold of Core-Feature to DEBUG (To receive some heart-beat LogEvents)")
    demo_logger.info("Note: The Demo_Minimal firmware emits a DEBUG LogEvent once a second; a heartbeat for demo.")
    device_proxy.core.evt_log.set_log_threshold(logging.DEBUG)
    provoke_some_log_events(device_proxy.router.transport)

    demo_logger.info("__________________________________________________________________________________________")
    demo_logger.info(f"Setting LogLevelThreshold of Core-Feature to WARNING")
    device_proxy.core.evt_log.set_log_threshold(logging.WARNING)
    provoke_some_log_events(device_proxy.router.transport)

    demo_logger.info("________________________________________________________________________________________")
    demo_logger.info(f"Setting LogLevelThreshold of Core-Feature to ERROR")
    device_proxy.core.evt_log.set_log_threshold(logging.ERROR)
    provoke_some_log_events(device_proxy.router.transport)

    demo_logger.info("________________________________________________________________________________________")
    demo_logger.info(f"Re-setting LogLevelThreshold of Core-Feature to INFO "
                     f"and checking whether device does proper trimming of LogLevelThreshold values")
    device_proxy.core.evt_log.set_log_threshold(logging.INFO)

    assert (device_proxy.core.prop_log_event_threshold.set(logging.DEBUG - 1) == logging.DEBUG)
    assert (device_proxy.core.prop_log_event_threshold.set(logging.DEBUG + 4) == logging.DEBUG)
    assert (device_proxy.core.prop_log_event_threshold.set(logging.DEBUG + 5) == logging.INFO)
    assert (device_proxy.core.prop_log_event_threshold.set(logging.CRITICAL + 1) == logging.CRITICAL)

    # Example of how numeric LogLevelThreshold values can be obtained as standard python names
    demo_logger.info("________________________________________________________________________________________")
    demo_logger.info(f"LogLevelThreshold of Core-Feature is currently set "
                     f"to {device_proxy.core.prop_log_event_threshold.get_value_name()}")

    device_proxy.router.close()


if __name__ == '__main__':
    showcase_logging()



import logging

logger = logging.getLogger(__name__)  # Logger-name: "hdcproto.transport.packetizer"


class Packetizer:
    """
    Packetizer of the HDC protocol.
    It transforms an incoming stream of bytes into individual messages and vice-versa.
    For details, please refer to the HDC-spec: https://github.com/kiksotik/hdc/blob/main/doc/spec/HDC-Spec.pdf
    """

    TERMINATOR = 0x1E  # "Record Separator" as defined by the ASCII standard
    MAX_PAYLOAD_SIZE = 0xFF  # Maximum payload that can be sent in a single package
    EMPTY_PACKET = bytearray([0, 0, TERMINATOR])  # payload_size=0 ; checksum=0 ; terminator

    incoming_raw_data_bytes: bytearray
    incoming_multi_package_message: bytearray
    reading_frame_error_count: int
    _received_messages: list[bytes]

    def __init__(self):
        self._received_messages = list()
        self.incoming_raw_data_bytes = bytearray()
        self.incoming_multi_package_message = bytearray()
        self.reading_frame_error_count = 0

    def clear(self):
        del self.incoming_raw_data_bytes[:]
        del self.incoming_multi_package_message[:]

    def data_received(self, data):
        """
        Received either additional bytes or nothing, which means that
        the timeout elapsed and thus any data burst is over.
        """

        if logger.isEnabledFor(logging.DEBUG):
            # Reminder: data.hex() is expensive and this is in the "hot-path" of the receiver loop!
            logger.debug("Received a chunk of %d bytes: [%s]", len(data), data.hex(sep=','))

        self.incoming_raw_data_bytes.extend(data)
        data_burst_is_over = (len(data) == 0)
        while self.incoming_raw_data_bytes:
            payload_length = self.incoming_raw_data_bytes[0]
            terminator_index = payload_length + 2
            if data_burst_is_over:
                logger.debug("Data-burst is over.")
            if terminator_index >= len(self.incoming_raw_data_bytes) and not data_burst_is_over:
                # Message is larger than the amount of data we currently have in the incoming_raw_data_bytes.
                # Wait for next data chunk or for a read-timeout that would mean that the data burst is over.
                logger.debug("Waiting for remainder of the message.")
                return
            if terminator_index < len(self.incoming_raw_data_bytes):
                if self.incoming_raw_data_bytes[terminator_index] == Packetizer.TERMINATOR:
                    # Uint8 sum of payload-bytes and two's complement checksum-byte must always yield zero
                    checksum = sum(self.incoming_raw_data_bytes[1:payload_length + 2]) & 0xFF
                    if checksum == 0x00:
                        payload = self.incoming_raw_data_bytes[1:payload_length + 1]
                        del self.incoming_raw_data_bytes[0:terminator_index + 1]
                        if self.incoming_multi_package_message or payload_length == self.MAX_PAYLOAD_SIZE:
                            self.incoming_multi_package_message.extend(payload)
                            if payload_length < self.MAX_PAYLOAD_SIZE:
                                self._received_messages.append(bytes(self.incoming_multi_package_message))
                                del self.incoming_multi_package_message[:]
                                logger.info("Unpacked one multi-packet message"
                                            "There's now %d messages in the queue.", len(self._received_messages))
                        else:
                            self._received_messages.append(bytes(payload))
                            logger.info("Unpacked one single-packet message. "
                                        "There's now %d messages in the queue.", len(self._received_messages))
                        continue
                    else:
                        logger.error("Incorrect checksum!")
                else:
                    logger.error("Incorrect terminator!")
            else:
                logger.error("Incorrect packet-size! "
                             "(Larger than the actual bytes received after completion of the burst of chunks.)")

            # ... otherwise it's very likely a reading-frame error!
            self.reading_frame_error_count += 1
            del self.incoming_raw_data_bytes[0:1]  # Skip first byte and try again.
            logger.error("Failed to unpack received bytes. Assuming a reading-frame error. "
                         "(RFE-counter: %d)", self.reading_frame_error_count)
            # Abort any ongoing multi-package message that we might have been receiving
            if self.incoming_multi_package_message:
                logger.warning("Aborting multi-package message reception, due to reading-frame-error.")
                del self.incoming_multi_package_message[:]
            # ToDo: Signal reading-frame error!
        else:
            if data_burst_is_over:
                logger.debug("Read attempt timed-out without receiving any data.",)

    def get_received_messages(self) -> list[bytes]:
        """
        Returns all messages received so far and forgets about them each time it's called.
        """
        tmp = self._received_messages
        self._received_messages = list()
        num_msg = len(tmp)
        if num_msg:
            logger.info("Handing over %d message(s). Queue is now empty.", num_msg)
        return tmp

    @staticmethod
    def compute_checksum(payload: bytes):
        """Returns the 8-bit two's complement checksum of the given block of bytes"""
        return (0xFF - sum(payload) + 1) & 0xFF

    @staticmethod
    def pack_message(message: bytes) -> list[bytes]:
        packets = list()

        # The following works for empty messages and also single and multi-packet messages
        multi_packet_payloads = (message[pos:pos + Packetizer.MAX_PAYLOAD_SIZE]
                                 for pos in range(0, len(message), Packetizer.MAX_PAYLOAD_SIZE))
        last_payload_size = Packetizer.MAX_PAYLOAD_SIZE  # Trick to produce an empty packet when the loop is skipped
        for payload in multi_packet_payloads:
            packet = bytearray()
            last_payload_size = len(payload)
            packet.append(last_payload_size)
            packet.extend(payload)
            packet.append(Packetizer.compute_checksum(payload))
            packet.append(Packetizer.TERMINATOR)
            packets.append(packet)

        if last_payload_size == Packetizer.MAX_PAYLOAD_SIZE:
            # Send empty packet to signal either:
            #    - ...the end of a multi-package message whose payload size is an exact multiple of 255 bytes.
            #    - ...or it might have been an empty message to begin with.
            packets.append(Packetizer.EMPTY_PACKET)

        logger.info("Packed a %d byte message into %d packets", len(message), len(packets))

        return packets
